#!/bin/bash
# flac2m4a.sh: Transcode FLAC audio files to .m4a audio files
# James A. Hillyerd
# $Revision: 24 $, $Date: 2006-10-04 19:10:00 -0700 (Wed, 04 Oct 2006) $
# Release: 1.1
#
# Usage: flac2m4a.sh <input-dir> <output-dir>
#
# Changelog:
#
#  Tues June 21 2011 Ron Olsen <ronolsen@comcast.net>
#  * Don't try to encode FLAC files with sample rate > 48 kHZ, since ffmpeg fails on these files.
#  * Check for errors on encoding, tagging, and adding cover art.
#  * Delete zero-length .m4a file which is created when a FLAC file fails to encode.
#  * Continue on an error instead of aborting.
#
#  Fri May 20 2011 Ron Olsen <ronolsen@comcast.net>
#  * Use <input-dir> and <output-dir> arguments
#
#  12/01/2009 - modified to trancode to m4a for the VortexBox project
#  01/05/2010 - added better support for Album art and AlbumArtist

### Begin Configuration
# The values below will be overridden by those in ~/.flac2mp3.conf if it
# exists

# Copy artwork (*.jpg) files? Set to "true" or "false"
COPY_ARTWORK="true"

# Where we store working data (including the decoded wav file)
TMPDIR="/tmp"

# Path to flac binary
FLAC="$(which flac)"

# Path to metaflac binary (part of flac distribution)
METAFLAC="$(which metaflac)"

# Path to ffmpeg binary 
FFMPEG="$(which ffmpeg)"

# Path to ffmpeg binary 
MP4TAGS="$(which mp4tags)"

# Maximum number of times we allow flac decode to fail per file
# before aborting
MAXFAILS=3

touch /etc/vortexbox/mirror-flac-aac.running

# Check that we have all the required binaries
if [ ! -x "$FLAC" ]; then
  echo "ERROR: flac not installed" >&2
  exit 1
fi

if [ ! -x "$METAFLAC" ]; then
  echo "ERROR: metaflac not installed" >&2
  exit 1
fi

if [ ! -x "$FFMPEG" ]; then
  echo "ERROR: ffmpeg not installed installing..." >&2
  yum -y --enablerepo=atrpms install ffmpeg --skip-broken
fi

if [ ! -x "$MP4TAGS" ]; then
  echo "ERROR: mp4v2 not installed installing..." >&2
  yum -y install mp4v2 libmp4v2 --skip-broken
fi


# Were we called correctly?
if [ -z "$1" -o -z "$2" ]; then
  echo "Usage: flac2m4a.sh <input-dir> <output-dir>" >&2
  exit 1
fi
NOW=`date +'%m/%d/%Y %H:%M:%S'`
function echolog {
echo "`date +'%m/%d/%Y %H:%M:%S'` - $1"
}

WORK="$TMPDIR/flac2mp3.$$"
ERR="$WORK/stderr.txt"
TAGS="$WORK/track.tags"

# Get absolute directory paths
INPUTDIR="$1"
OUTPUTDIR="$2"

#INPUTDIR="/storage/music/flac"
#OUTPUTDIR="/storage/music/m4a"

echo ""
echo ""
echolog "Starting FLAC to Apple Lossless (.m4a) mirror..."
echo ""


# Setup work directory
rm -rf "$WORK"
if [ -e "$WORK" ]; then
  echo "Couldn't delete $WORK"
  exit 1
fi
mkdir "$WORK"
if [ ! -d "$WORK" ]; then
  echo "Couldn't create $WORK"
  exit 1
fi

# Get input file list
cd "$INPUTDIR"
OLDIFS="$IFS"
IFS=$'\n'
for filepath in $(find . -type f -name '*.flac' -print \
    | sort | sed -e 's/^\.\///' -e 's/\.flac$//'); do
  IFS="$OLDIFS"
  filedir="$(dirname "$filepath")"
  filename="$(basename "$filepath")"
  if [ ! -d "$OUTPUTDIR/$filedir" ]; then
    mkdir -p "$OUTPUTDIR/$filedir"
  fi
  
  if [ ! -f "$OUTPUTDIR/$filepath.m4a" ]; then

    # Clear tags
    rm -f "$TAGS"
    ARTIST=
    ALBUM=
    ALBUMARTIST=
    TITLE=
    DATE=
    GENRE=
    COMPOSER=
    DISCNUMBER=
    TRACKNUMBER=

    # Get tags
    $METAFLAC --export-tags-to="$TAGS" "$INPUTDIR/$filepath.flac"
    sed -i -e 'h
        s/^[^=]*=//
        s/"/\\"/g
        s/\$/\\$/g
        s/^.*$/"&"/
        x
        s/=.*$//
        y/abcdefghijklmnopqrstuvwxyz/ABCDEFGHIJKLMNOPQRSTUVWXYZ/
        G
        s/\n/=/' "$TAGS"
    source "$TAGS"
	echolog "########################################"
	echo "Encoding $TITLE"
	echo "Artist: $ARTIST"
	echo "Album: $ALBUM ($DATE)"
	echo "AlbumArtist: $ALBUMARTIST"
        echo "Composer: $COMPOSER"
        echo "DiscNumber: $DISCNUMBER"
	echo "Track#: $TRACKNUMBER"
	echo "Genre: $GENRE"
	echo ""
	
    # Encode M4a
      # Skip high-resolution FLAC files (sample rate > 48 kHZ) which cannot be encoded to .m4a by ffmpeg
      sample_rate=$($METAFLAC --show-sample-rate "$filepath.flac")
      if [ $sample_rate -gt 48000 ]
      then
         echo "WARNING: Sample rate $sample_rate > 48 kHZ, cannot encode to .m4a, skipping"
         echo ""
         continue
      fi
      ffmpeg -i "$filepath.flac" -acodec alac "$OUTPUTDIR/$filepath.m4a" >/dev/null 2>&1
      if (( $? )); then
        echo "ERROR: encoding failed, continuing"
        echo ""
        rm -f "$OUTPUTDIR/$filepath.m4a"
        continue
      fi

      if [ "`echo $DISCNUMBER | egrep ^[[:digit:]]+$`" = "" ]; then
        mp4tags -s "$TITLE" --track "$TRACKNUMBER" -g "$GENRE" -year "$DATE" -a "$ARTIST" -A "$ALBUM" -R "$ALBUMARTIST" -w "$COMPOSER" "$OUTPUTDIR/$filepath.m4a"
      else
        mp4tags -s "$TITLE" --track "$TRACKNUMBER" -d "$DISCNUMBER" -g "$GENRE" -year "$DATE" -a "$ARTIST" -A "$ALBUM" -R "$ALBUMARTIST" -w "$COMPOSER" "$OUTPUTDIR/$filepath.m4a"
      fi
      if (( $? )); then
        echo "ERROR: tagging failed, continuing"
        echo ""
        rm -f "$OUTPUTDIR/$filepath.m4a"
        continue
      fi
	
      if [ -f "$INPUTDIR/$filedir/cover.jpg" ]; then         
         mp4art --add "$INPUTDIR/$filedir/cover.jpg" "$OUTPUTDIR/$filepath.m4a"
         if (( $? )); then
           echo "ERROR: Adding cover art failed, continuing"
           echo ""
           #rm -f "$OUTPUTDIR/$filepath.m4a"
           continue
         fi
      fi
    echo ""
  fi
done

# Clean up
rm -rf "$WORK"

echo ""
echolog "FLAC to Apple Lossless (.m4a) mirror complete."
echo ""
rm /etc/vortexbox/mirror-flac-aac.running
exit
