#!/bin/bash
# flac2mp3.sh: Transcode FLAC audio files to MP3 audio files
# James A. Hillyerd
# $Revision: 24 $, $Date: 2006-10-04 19:10:00 -0700 (Wed, 04 Oct 2006) $
# Release: 1.1
#
# Usage: flac2mp3 <input-dir> <output-dir>
#
#  03/22/2010 - modified to trancode to wav for the VortexBox project
#  06/30/2010 - updated to support tagging wav files

### Begin Configuration
# The values below will be overridden by those in ~/.flac2mp3.conf if it
# exists

# Copy artwork (*.jpg) files? Set to "true" or "false"
COPY_ARTWORK="true"

# Where we store working data (including the decoded wav file)
TMPDIR="/tmp"

# Path to flac binary
FLAC="$(which flac)"

# Path to metaflac binary (part of flac distribution)
METAFLAC="$(which metaflac)"

# Maximum number of times we allow flac decode to fail per file
# before aborting
MAXFAILS=3

touch /etc/vortexbox/mirror-flac-wav.running

# Check that we have all the required binaries
if [ ! -x "$FLAC" ]; then
  echo "ERROR: flac not installed" >&2
  exit 1
fi

if [ ! -x "$METAFLAC" ]; then
  echo "ERROR: metaflac not installed" >&2
  exit 1
fi

NOW=`date +'%m/%d/%Y %H:%M:%S'`
function echolog {
echo "`date +'%m/%d/%Y %H:%M:%S'` - $1"
}

WORK="$TMPDIR/flac2wav.$$"
ERR="$WORK/stderr.txt"
TAGS="$WORK/track.tags"

# Get absolute directory paths
INPUTDIR="$1"
OUTPUTDIR="$2"

INPUTDIR="/storage/music/flac"
OUTPUTDIR="/storage/music/pcm"

echo ""
echo ""
echolog "Starting FLAC to wav mirror..."
echo ""


# Setup work directory
rm -rf "$WORK"
if [ -e "$WORK" ]; then
  echo "Couldn't delete $WORK"
  exit 1
fi
mkdir "$WORK"
if [ ! -d "$WORK" ]; then
  echo "Couldn't create $WORK"
  exit 1
fi

# Get input file list
cd "$INPUTDIR"
OLDIFS="$IFS"
IFS=$'\n'
for filepath in $(find . -type f -name '*.flac' -print \
    | sort | sed -e 's/^\.\///' -e 's/\.flac$//'); do
  IFS="$OLDIFS"
  filedir="$(dirname "$filepath")"
  filename="$(basename "$filepath")"
  if [ ! -d "$OUTPUTDIR/$filedir" ]; then
    mkdir -p "$OUTPUTDIR/$filedir"
  fi
  
  if [ ! -f "$OUTPUTDIR/$filepath.wav" ]; then

    # Clear tags
    rm -f "$TAGS"
    ARTIST=
    ALBUM=
    ALBUMARTIST=
    TITLE=
    DATE=
    GENRE=
    COMPOSER=
    DISCNUMBER=
    TRACKNUMBER=

    # Get tags
    $METAFLAC --export-tags-to="$TAGS" "$INPUTDIR/$filepath.flac"
    sed -i -e 'h
        s/^[^=]*=//
        s/"/\\"/g
        s/\$/\\$/g
        s/^.*$/"&"/
        x
        s/=.*$//
        y/abcdefghijklmnopqrstuvwxyz/ABCDEFGHIJKLMNOPQRSTUVWXYZ/
        G
        s/\n/=/' "$TAGS"
    source "$TAGS"
	echolog "########################################"
	echo "Encoding $TITLE"
	echo "Artist: $ARTIST"
	echo "Album: $ALBUM ($DATE)"
	echo "AlbumArtist: $ALBUMARTIST"
        echo "Composer: $COMPOSER"
        echo "DiscNumber: $DISCNUMBER"
	echo "Track#: $TRACKNUMBER"
	echo "Genre: $GENRE"
	echo ""
	
    # Encode wav
	flac -d "$filepath.flac" -o "$OUTPUTDIR/$filepath.wav" >/dev/null 2>&1
      if [ "`echo $DISCNUMBER | egrep ^[[:digit:]]+$`" = "" ]; then
        /opt/vortexbox/wavtags -s "$TITLE" -t "$TRACKNUMBER" -g "$GENRE" -y "$DATE" -a "$ARTIST" -A "$ALBUM" -r "$ALBUMARTIST" -w "$COMPOSER" "$OUTPUTDIR/$filepath.wav"
      else
        /opt/vortexbox/wavtags -s "$TITLE" -t "$TRACKNUMBER" -d "$DISCNUMBER" -g "$GENRE" -y "$DATE" -a "$ARTIST" -A "$ALBUM" -r "$ALBUMARTIST" -w "$COMPOSER" "$OUTPUTDIR/$filepath.wav"
     fi
      if (( $? )); then
        echo ""
        echo "Encode failed, aborting"
        exit 1
      fi

    echo ""
    echo ""

  fi
done

cd /storage/music/flac
find . -name cover.jpg -type f -execdir cp cover.jpg folder.jpg \;
rsync -av --include='*/' --include='cover.jpg' --include='folder.jpg' --include='*.txt'  --exclude='*' /storage/music/flac/ /storage/music/pcm/


# Clean up
rm -rf "$WORK"

echo ""
echo ""
echolog "FLAC to WAV mirror complete."
echo ""
rm /etc/vortexbox/mirror-flac-wav.running
exit
